/*
 *  diskdrvr.h  "Public" Disk Driver Definitions
 *
 *  Copyright (C) by Quantum Software Systems 1990.  All rights reserved.
 */
#ifndef __DISKDRVR_H_INCLUDED
#define __DISKDRVR_H_INCLUDED

#pragma pack(1);

#ifndef _DISK_H_INCLUDED
 #include <sys/disk.h>
#endif

#define _FSYS_NAME  "qnx/fsys"
#define _MAX_BLOCK_OPS  127 /*  Largest atomic block_read or block_write    */

/*
 *  I/O modes
 */
#define _READ_BLK       0
#define _WRITE_BLK      1

/*
 *  Disk driver jump table indices
 */
#define _DRIVER_TBL_SIZE        16

#define _DRIVER_INITIALIZE      0   /*  Initialize driver/controller        */
#define _DRIVER_TERMINATE       1   /*  Driver shutdown                     */
#define _DRIVER_UNIT_INIT       2   /*  Open (and lock) a disk for use      */
#define _DRIVER_UNIT_TERM       3   /*  Close (and unlock) a device         */
#define _DRIVER_IO              4   /*  Read / Write media                  */
#define _DRIVER_CONTROL         5   /*  Everything else                     */
#define _DRIVER_CONTROL_MASK    6   /*  Effect of control (ioctl) request   */
#define _DRIVER_IO_FINISH       7   /*  Called after proxy to complete      */
                                    /*  I/O or CONTROL requests             */
#define _DRIVER_UNIT_OPEN       8   /*  Called everytime unit is opened     */
                                    /*  Mounting a filesystem is an open    */
#define _DRIVER_UNIT_CLOSE      9   /*  Called after last close only        */

/*
 *  Control (ioctl) mask bits.
 *  These define the effect of a control (ioctl) request.
 *  If the _DRIVER_CONTROL_MASK entry point is NULL but _DRIVER_CONTROL exists,
 *  assume the mask is 0xFFFF for backwards compatibility.
 */
#define _DRVR_CTRL_SUPPORTED    0x0001  /*  Request is supported            */
#define _DRVR_CTRL_PRIVILIGED   0x0002  /*  Request is priviliged (uid == 0)*/
#define _DRVR_CTRL_DESTRUCTIVE  0x0004  /*  Will change contents of disk,   */
                                        /*  invalidating cache (e.g. format)*/
#define _DRVR_CTRL_RECONFIG     0x0008  /*  Will change disk size           */
#define _DRVR_CTRL_ASYNCH       0x0010  /*  Must wait for device to be idle */
                                        /*  before issuing call, and wait   */
                                        /*  for proxy to indicate request   */
                                        /*  is complete                     */

/*
 *  Structure of a file system cache block entry
 */
struct _block {
    unsigned short   b_next_off,        /*  Offset of next buffer in list   */
                     b_reserved;
    long             b_block;           /*  Block number                    */
    char __far      *b_ptr;             /*  Buffer data pointer             */
    } ;

/*
 *  Physical disk information
 *
 *  A driver will have an array of these structures;
 *  one of these for each physical disk a driver/controller supports.
 *
 *  So long as the d_num_blks field is filled correctly,
 *  the other size fields (e.g. d_tracks, d_heads, etc.), may be left empty
 *  (e.g. for SCSI or tapes, where #heads, #tracks, etc. make no sense)
 *  The exception is d_type == _TAPE, where the size isn't known.
 *  In that case, you can leave d_num_blks == 0.
 */

struct _disk {
    long                    d_num_blks; /*  Size of disk in blocks      */
    unsigned short          d_type,     /*  Disk type                   */
                            d_cyls,     /*  Number of tracks            */
                            d_sctr_cyl, /*  Sectors per cylinder        */
                            d_heads,    /*  Number of heads             */
                            d_sctr_trk, /*  Sectors per track           */
                            d_sctr_base,/*  Sector base (0 or 1)        */
                            d_error;    /*  Open/Close error number     */
    volatile unsigned short d_busy,     /*  Disk busy performing I/O    */
                            d_nblks,    /*  Number of blocks I/O'd      */
                            d_retval,   /*  Return value for an I/O call*/
                            d_dma_seg;  /*  Zero if data put in cache   */
                                        /*  If non-zero, Fsys must copy */
                                        /*  to cache after proxy        */
    char                    d_prefix[8];/*  Device name prefix          */
                                        /*  May be same as driver name  */
    unsigned short          d_stat,     /*  Unit status type            */
                            d_reserved[6];  /*  Reserved for future use */
    } ;


/*
 *  Driver return values.
 *  (Driver will map hardware results to a standard return value for FSYS.)
 *  Return values with the top bit set are *REAL* bad!
 */
#define _DRVR_OK            0       /*  Request was 100% successful.        */
#define _DRVR_PART_IO       0x0001  /*  d_nblks were OK, but next is bad.   */
#define _DRVR_RESERVED      0x0002  /*  We have future plans for this one.  */

#define _DRVR_UNSPEC        -1      /*  Unspecified error.                  */
#define _DRVR_ERROR         0x8001  /*  Nothing worked!                     */
#define _DRVR_DISK_FAILURE  0x8002  /*  Driver can't talk with hardware.    */
                                    /*  For floppy and removable, possibly  */
                                    /*  disk has been removed from drive.   */
#define _DRVR_EOT           0x8003  /*  Driver has hit end-of-tape marker.  */
                                    /*  Indicate EOF to app. for read       */
                                    /*  and ENOSPC error for write.         */

/*
 *  Unit status bits.
 */
#define _DRVR_RDONLY        0x0001  /*  Unit is write protected             */
#define _DRVR_NO_MEDIA      0x0002  /*  Someone removed the media!          */

/*
 *  Driver information flags.
 */
#define _DRVR_DMA           0x0001  /*  Driver uses DMA.                    */
#define _DRVR_DMA_DUMB      0x0002  /*  Dumb DMA implementation.            */
                                    /*  Cannot be run at same time          */
                                    /*  as bus mastering DMA, e.g. floppy.  */

#pragma pack();

#endif
