
#ifdef __USAGE
%C	[-a controller] [-i interrupt] [-m maxsize] [-n name]
Options:
 -a controller  The controller address (default is 3f0 hex).
 -i interrupt   The interrupt level used by the controller (defult is 6).
 -m maxsize     The maximum number of blocks that can be transferred from the
                diskette to memory in one operation (default is 36).
 -n name        The name of the block special device associated with the
                floppy. The default is 'fd' e.g. /dev/fd0, /dev/fd1
#endif

/*
	Qnx 4.0 Floppy Disk Driver.


	
	Description:

	This driver presents the floppy drive interface to the filesystem, 
	supporting multi sector io (up to num sectors per cylinder), automatic 
	drive resizing, and diskette formatting.

	The logic for the driver is controlled by a 'action table', which 
	encodes a finite state machine.   The table contains pointers to
	functions, indexed by [current_state][input_event], where input_event
	may be one of (Interrupt, IORequest, TimeOut).  Due to the nature of
	the interface, several states must be 'timed' and requests aborted if
	not satisfied in sufficient time.  When modifying this driver, the best
	approach is to understand the state table, and make modifications there,
	rather than trying to overload a state with more functions.   

	The driver is intended to be separate from the hardware interface,
	which is implimented in a separate module.  However, due to the nature
	of the device, some of the logic is inexonarably tied to the NEC 765:
	INTEL/8272 floppy disk controller.

	Transition table:
		The left hand side, is the state, the right hand list are the
		states reachable in 1 event from the state.

    ST_START	:	ST_VIRGIN, ST_IDLE, ST_DEAD
	ST_IDLE	 	:	ST_START, ST_SPIN
	ST_SPIN	 	:	ST_START, ST_SEEK
	ST_SEEK		:	ST_START, ST_XFER, ST_SEEK, ST_READY
	ST_XFER		:	ST_START, ST_READY, ST_SEEK, ST_SPECIFY, ST_IDLE
	ST_READY	:	ST_START, ST_SEEK, ST_IDLE
	ST_SPECIFY	:	ST_START, ST_SEEK, ST_SPECIFY
	ST_SPECIFY2	:	....
	ST_FMT_TRK	:	ST_START, ST_IDLE, ST_READY
 *should* never happen 
	ST_DEAD		:	ST_DEAD.
	
	ST_START:
		The floppy is untouched.  
		On an IORequest, the drive is reset.  If the drive responds with an
		interrupt within 200 msec,it moves to state "ST_IDLE".  
		If it does not respond in this period,it moves to state "ST_DEAD".
		Most states, on a sequence error (ie. get a timer-interrupt when one 
		is not 'possible'), will cause a move to 'virgin'.  This is relatively
		safe, as the first operation at ST_START is 'reset'.

	ST_IDLE:
		The floppy is ready for an IO Request.
		When an IORequest is specified (or follow-through from ST_START), the
		motor is turned on and driver parameters specified.  It moves to state
		ST_SPIN until a timer expires (600 msec or 1000 msec, depending upon
		the request).

	ST_SPIN:
		The floppy has 'spun up'.
		When a TimeOut event is received, a seek command is issued, and the
		state moves to "ST_SEEK".  Any other event causes an error, and the
		state moves to "ST_START".

	ST_SEEK:
		The seek command has completed.
		When an Interrupt is received, the device status is queried.  If the
		seek was successful, the data transfer command is issued, and the state
		moves to either ST_XFER or ST_FMT_TRK.  
		This command is protected by a timeout, to prevent
		the drive from never returning if it is not ready. If the seek failed,
		the seek command is reissued if the retry count is below MAX_RETRY.  
		Otherwise, the IO fails, a timer is set, and the state moves to ST_READY.

	ST_XFER:
		The data transfer has completed.
		When an Interrupt is received, the timer is cancelled, and the result
		code examined.  If the result code shows data transfer occurred, then
		the number of blocks successfully transfered is examined, and the
		IORequest completes sucessfully.  If the first block of the disk is 
		observed, the drive examines known areas and sets the drive parameters
		(h, t, s) if the disk is a known  qnx4.0, qnx2.XX, or DOS X.XX format.
		A timer is set, and the state then moves to ST_READY.

		If the result code shows an error, the the retry logic is executed.
		On certain retry counts, the drive parameters are changed to adapt to
		the radically different 360/1.2M style drives.  If, after MAX_RETRY 
		has expired, the IO request is failed,a timer is set, and the state 
		moves to ST_READY.  On some of the retry counts, a recalibrate
		is issued and the state moves to 'ST_SPECIFY'.

		If a Timeout occurs, IO request is failed, the drive moves to state
		ST_START, the motors are shut off,

	ST_READY:
		Drive is ready for a new command.
		If an IORequest occurs, the timer is removed, an appropriate
		seek request is issued, and the drive moves to ST_SEEK.
		If a TimeOut occurs, the motor is shut off, and the drive moves
		to ST_IDLE.

	ST_SPECIFY:
		The drive is recalibrating on an error.
		If an Interrupt occurs, the status is checked.  If it fails,
		the recalibrate is re-issued.  If successful, the seek is issued,
		and the drive moves to state "ST_SEEK".
		Any other event causes it to move to "ST_START".

	ST_FMT_TRK:
		The drive has completed formatting a track.
		If an Interrupt is recieved, the state moves to ST_READY, and the
		command completes, either sucessfully or fails.
		If a TimeOut is received, the state moves to S_IDLE, and the
		command fails.

	ST_DEAD:
		The drive is DEAD.
		An IORequest results in an immediate failure.
		Any other event is considered a 'no-op'.



	ToDo:
		Clean up format command, allow driver query protocol.

		A future may be a 'raw-io' interface.  My thought is something like
		this:
			nmsgs <msg>*
			msg :: opcode, len, data{len,len}
			opcode ::	NEC765_command
						CLOCK_command
						DMA_command
						MOTOR_command
						.....
		The idea being to allow user programs to send floppy commands directly
		(ala scsi), and get responses back.  They could provide dma buffers
		directly in the opcode, which would be setup and the command executed.
		Could this support QIC-02 interfaces????


	Steve McPolin.

*/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <i86.h>

#include <sys/name.h>
#include <sys/irqinfo.h>
#include <sys/kernel.h>
#include <sys/proxy.h>
#include <sys/seginfo.h>
#include <diskio.h>
#include "floppy.h"




#ifndef DEBUGGING
#define	DEBUGGING 0
#endif

#include <sys/osinfo.h>
struct _osinfo _osdata;

char driver_name[_DNAME_PREFIX+1] = FLOPPY_DISK_PREFIX;


struct	_sysmsg_version_reply	version_stamp = {
		"Floppy",
		__DATE__,
		400,
		'B',
		0
	};



/*	makes it easier to derive a map from the executable:
	however, compiler can generate better code if it knows
	the scope of a variable is restricted.
*/

#define	LOCAL	static


/*

	data shared between interrupt handler & call level.
	Parameters, etc.. from call level are remembered here.


	Although the floppy controller allows some overlap between
	instructions, this overlap is not very usefull, as the motor
	is controlled by mutually exlusive register bits.

	Thus the driver_io function must determine whether there is any
	current activity on any floppy.  If there is, return -1.
	If the last/accessed| ready floppy is not this one, turn it's motor 
	off and move to state IDLE.
	
*/

LOCAL	unsigned	Function;		/* Function interrupt handler is doing	*/
LOCAL	long		FirstBlock;		/* First block to retrieve */
LOCAL	unsigned	NumBlocks;		/* Number of blocks yet to transfer		*/
LOCAL	int			DriveNo = 0;	/* Disk number */

/*	This is required for fsys */
LOCAL	BLOCK __far *CachePtr;		/* Linked list of blks read/written		*/
LOCAL	BDATA		CurRequest;	/* Current Request being Served         */



/*
	These areas are for the binding between the filesystem and the driver,
	they are 'relative' constants, that is they can only be redefined at
	startup.

*/

LOCAL	int			Ndrives = MAX_NDRIVE;	/*	init to Maximum. */
LOCAL	pid_t		NotifyPid;			/* Proxy for int handler to use	*/
LOCAL	pid_t		FsysPid;
LOCAL	unsigned	MaxIO				= MAX_IO,
					ControlInterrupt	= CTL_INTERRUPT;


LOCAL	int		base_address = CTL_ADDRESS;


/*	These are the cookies for the interrupt handler and timer handler
	respect.
*/

LOCAL	int		int_id = 0,
				tint_id = 0;

int	multi_track = 1;


/*	Current state of the driver	*/
int FdcState = ST_START;

/*
	This buffer is defined when the driver is initialed.
*/

LOCAL	dma_flags = 0;
LOCAL	dma_channel = 2;
LOCAL	struct dma_buffer	iobuffer;	



/* Disk parameter tables	*/
#if 0

Something NEW:

I am changing the drive parameter tables to hopefully fullfill the
requirements for the bizarre combinations of floppy drive parameters

There is a table, of this structure:

struct	fparm {
	short int	bios_type;
	short int	read_gap;
	short int	fmt_gap;
	short int	head_load;
	short int	head_unload;
	short int	pll_clock;
	short int	step_rate;
	short int	dstep;
};


struct	fparm	bios_types[] = {
/*360 drive */	{	FD_360,   42, 47,   7,   15, FD_300KBS, 13, 0 },
/*1.2 drive */	{	FD_360,   35, 47,   7,  15, FD_300KBS, 12, 1 },
/*1.2 drive */	{	FD_1200,  27, 108,  7,   15, FD_500KBS, 10, 0 }, 
/*144/720  */	{	FD_720,   42, 47,   7,   15, FD_250KBS, 13, 0 }, 
/*144 drive */	{	FD_144,   27, 108,  7,   15, FD_500KBS, 10, 0 }
	};

struct	device {
	short int	bios_type;
	short int	cparm;
	short int	mparm;
	short int	locked;
	DISK		*dptr;
	short int	fparm_list[MAX_CONF];
};

Where MAX_CONF is the maximum number of setups any drive can support.
bios_type is the type of drive, as determined by the bios.
cparm	  is the index into fparm_list for the current setup.
mparm	  is the maximum entries for this drive type.
locked	  is whether the drive is currently locked or not.
fparm_list are indices into the setup table.

	When the first block of the disk is read, the drive parameters are derived
	and the drive is "locked".   The drive remains locked until another read of
	block 0 is made, or an IOCTL call is made to set the drive parameters.
	Setting the drive parameters 


#endif	
	
LOCAL DISK	disk[MAX_NDRIVE];

typedef	struct {
		int	type;			/*	what cmos says it is.... */
		int	locked;			/*	have we locked onto this drive yet? */
		int	cur;			/*	0 or 1 -- which are we trying? */

		int	clock[2];		/*	transfer rate	*/
		int	st_rate[2];		/*	step rate */
		int	rd_gap[2];		/*	read gap  */
		int	dstep[2];		/*	double step {0,1}	*/
		int	fmt_gap[2];		/*	format gap */

		int	head_load;		/*	head load time	*/
		int	head_unload;	/*	head unload time */
		/*	any other weenie values here.	*/
	}	FLOPPY_PARMS;

LOCAL	FLOPPY_PARMS	DriveParms[MAX_NDRIVE];
LOCAL	FLOPPY_PARMS	*cur_disk;

LOCAL	struct fd_fmt_trk	__far *fmttrk;



/*
 *	Driver functions, a jump table through which fsys enters the driver code
 *
 *	Array of pointers to driver functions.
 */


pid_t __far	int_handler(void);
pid_t __far int_timer(void);


/*
 *  Driver entry points.
 */

int __far drvinit(pid_t driver_proxy, short tabsize, unsigned __far *cachetab);
int __far drvterm(unsigned flags, int rsvd1, int rsvd2, int rsvd3);
int __far drvopen(int disknum, int flags, int rsvd1, int rsvd2);
int __far drvclose(int disknum,int rsvd1, int rsvd2, int rsvd3);
int __far drvio(unsigned disknum, unsigned mode,
                unsigned num_blocks, unsigned rsvd,
                struct _block __far *bp);
int __far drvctl(unsigned disknum, unsigned request,
				 unsigned rsvd1, unsigned rsvd2,
				 void __far *cdata);
int __far drvctmask(unsigned disknum, unsigned request,
				unsigned rsvd1, unsigned rsvd2);
				
int __far driver_init( pid_t driver_proxy );
int __far driver_terminate( void );
int __far driver_disk_open( int disknum, int pflag );
int __far driver_disk_close( int disknum );
int __far driver_io(unsigned __disknum, unsigned write_flag,
                        unsigned num_blocks, struct _block __far *bp );
int __far driver_control( unsigned disknum, unsigned request,
						  unsigned reserve,	void __far *cdata);
                    
int (*drvr_entry[_DRIVER_TBL_SIZE])() = {
			drvinit,
			drvterm,
			drvopen,
			drvclose,
			drvio,
			drvctl,
			drvctmask
		};


char *pname = "floppy";

int	debug_sel = -1;
short int __far *videoptr;
int	video_wrap = 80*15;
int	video_ind =5*80;

#if DEBUGGING
#define	VPUTC(_c) ((videoptr[video_ind] = (_c)|0x700), video_ind = video_ind < video_wrap ? video_ind+1 : 0)
#else
#define	VPUTC(_c)
#endif

int vputc(int c)
{
#if DEBUGGING
	return VPUTC(c);
#else
	return c;
#endif
}

main(int argc, char **argv) 
{
long	stack[STACK_SIZE];
void	*init_sptr = (void *)(stack + STACK_SIZE-1);

	pname = *argv;

	if ((Ndrives = process_args(argc, argv)) == 0) {
		fputs("No floppy drives detected, check setup\n",stderr);
		return EXIT_SUCCESS;
	}
	get_cpu_speed();
	fdc_init(base_address);
	if ((FsysPid = qnx_name_locate(getnid(),_FSYS_NAME,0,0)) == -1) {
		perror(_FSYS_NAME);
		fputs("Could not locate filesystem\n",stderr);
	}
#if DEBUGGING
	if (debug_sel == -1) {	/*	not a valid selector! */
		qnx_osinfo(0, &_osdata);
		debug_sel = (_osdata.primary_monitor & 001) ? MONO_DISP : COLOUR_DISP;
	}
	printf("enabling debug at selector %x\n",_osdata.machinesel);
	videoptr = MK_FP(debug_sel,video_ind);
	set_display(videoptr,10,65);
#endif
	FdcState = ST_START;

	if (fsys_mount_driver(FsysPid,Ndrives,disk,drvr_entry,
			MaxIO,MAX_PARALLEL,init_sptr,&version_stamp) == -1) {
		perror("mount");
		fputs("attach filesystem failed\n",stderr);
		return EXIT_FAILURE;
	}
	return EXIT_SUCCESS;
}




process_args(int argc, char **argv) 
{
extern int optind;
extern char *optarg;
int		opt;

	while ((opt=getopt(argc,argv,"a:i:m:n:**V:")) != EOF) {
		switch( opt ) {
		case 'a':		/*	Control address */
			base_address = strtol(optarg,NULL,16);
			break;
		case 'i':		/*	IRQ level */
			ControlInterrupt = strtol(optarg,NULL,0);
			break;
		case 'm':
			MaxIO = strtol(optarg,NULL,0);
			if (MaxIO >= 126)		/*	64k seg... */
				MaxIO = 1;
			break;
		case 'n':
			if (strlen(optarg) < 1){
				fprintf(stderr,"name too short\n");
				exit(EXIT_FAILURE);
			}
			if (strlen(optarg) > sizeof(driver_name)-1) {
				fprintf(stderr,"name too long\n");
				exit(EXIT_FAILURE);
			}
			strcpy(driver_name, optarg);
			break;
		case	'V':
			debug_sel = strtol(optarg,NULL,16);
			break;
		default:
			return 0;
			break;
		}
	}
	if (get_floppy_parms(disk+0,0) == -1) {
		return 0;
	}
	if (get_floppy_parms(disk+1,1) == -1) {
		return 1;
	}
	return 2;
}



/*
 *  Init the driver
*
 *	This provides an opportunity to perform device specific configuration
 *  while running as part of the file system.  Typically, allocating 
 *  buffers (dma), reseting the controller, squirreling away who fsys is,
 *  installing interrupt handlers.   Note that the latter *must* be done
 *  while in the context of the file-system if driver is not an actual
 *  process.  Also, the Code Segment stored must be relative to the filesystem,
 *  not the process (process may go away!)
 */

int __far drvinit(pid_t proxy, short tabsize, unsigned __far *cachetab)
{
	if (tabsize || cachetab)
		;
		
	NotifyPid = proxy;

	if (alloc_dma_buf(&iobuffer, FsysPid, MaxIO * _BLOCK_SIZE)) {
		return -1;
	}
{
unsigned long saddr;
unsigned long taddr;

	taddr = (saddr=iobuffer.phys_addr) + MaxIO * _BLOCK_SIZE;
	if ((taddr & 0xffff) < (saddr & 0xffff)) { 
		/*	the segment wrapped.... 
			get the most significant part,
			and limit MaxIO to that value / _BLOCK_SIZE,
			so DMA can NEVER exceed a 64K chunk....
		*/	
		taddr &= ~ 0xffff;
		MaxIO = (taddr-saddr)/_BLOCK_SIZE;
	}
}

	if ((tint_id = qnx_hint_attach(-1, MK_FP(my_cs(), &int_timer), my_ds())) == -1) {
		qnx_segment_free(FP_SEG(iobuffer.bufp));
		return -1;
	}
	if ((int_id = qnx_hint_attach(ControlInterrupt,
				MK_FP( my_cs(), &int_handler ),	my_ds())) == -1) {
		qnx_segment_free(FP_SEG(iobuffer.bufp));
		qnx_hint_detach(tint_id);
		return -1;
	}
	fdc_reset();
	return 0;
}

int __far drvterm(unsigned flags, int rsvd1, int rsvd2, int rsvd3)
{
	if (flags || rsvd1 || rsvd2 || rsvd3)
		;
	qnx_hint_detach(int_id);
	qnx_hint_detach(tint_id);
	qnx_segment_free(FP_SEG(iobuffer.bufp));
	return 0;
}




int	__far	drvopen(int disknum, int flags, int rsvd1, int rsvd2)
{
DISK *dp;

	if (flags || rsvd1 || rsvd2)
		;

	if (disknum >= MAX_NDRIVE) {
		return -1;
	}

	(dp=disk+disknum)->d_busy = 1;	/* The drive is busy while we open it			*/
	if (dp->d_type == _UNMOUNTED) {
		dp->d_retval = _DRVR_UNSPEC;
		dp->d_busy = 0;
		return dp->d_retval;
	}
	strcpy(dp->d_prefix,driver_name);
	dp->d_retval = 0;
    dp->d_busy = 0;
	return	dp->d_retval;
}

/*
 * driver_disk_close
 * 
 *  Mark stored driver information as invalid and continue.
 *
*/


int __far drvclose(int disknum,int rsvd1, int rsvd2, int rsvd3)
{
	if (rsvd1 || rsvd2 || rsvd3)
		;
	memset(disk+disknum,0,sizeof disk[disknum]);
	return 0;
}
	
/*

This table encodes the action and some of the logic for the floppy handlers.
Each function performs a small action, then 

*/

pid_t	reset_dev(), /* device is in unknown state, apply reset and set timer */
		init_dev(),	/* device is reset clear timer and init_io() */
		init_io(),	/* device is idle, start it up, wait for a timer */ 
		absorb_int(),	/* ignore interrupts from device at power-up */
		seek_drive(),  /* position head at proper cylinder */
		io_start(),	/* head is positioned, issue IO request */ 
		io_done(),	/* device has responded to io request */
		xfer_to(),	/* timer expired before device responded */
		fmt_done(),	/* device has responded to a format request */
	 	fmt_to(),	/* timer expired before device responded */
		ready_io(),	/* device is ready, issue a new io request */ 
		dev_idle(),	/* timer has elapsed since last io, turn off motor */
		retry(),	/* recalibrate phase of retry */ 
		retry2(),	/* final recalibrate phase    */
	
		io_trap(),		int_trap(),		timer_trap();
		



const pid_t	(*action[][MAX_EVENT])() = {

					/*  IORequest   TimeOut     Interrupt    */

/*ST_START  */		{	reset_dev,	timer_trap,	absorb_int },
/*ST_RESET */		{	io_trap,	timer_trap,	init_dev   },
/*ST_IDLE    */		{	init_io,	timer_trap,	int_trap   },
/*ST_SPIN    */		{	io_trap,	seek_drive,	int_trap   },
/*ST_SEEK    */		{	io_trap,	timer_trap,	io_start   },
/*ST_XFER    */		{	io_trap,	xfer_to,	io_done    },
/*ST_READY   */		{	ready_io,   dev_idle,   absorb_int },
/*ST_SEPCIFY */		{	io_trap,	timer_trap,	retry      },
/*ST_SPECIFY2*/		{	io_trap,	timer_trap,	retry2     },
/*ST_FMT_TRK */		{	io_trap,	fmt_to,		fmt_done   },
/*ST_RTRAP	 */		{	io_trap,	io_trap,	io_trap		},
/*ST_ITRAP	 */		{	int_trap,	int_trap,	int_trap	},
/*ST_TTRAP	*/		{	timer_trap,	timer_trap,	timer_trap	}
		};

/* 0a 1c 3b 4c 9c 6b
	reset_dev
	init_dev
	seek_drive
	io_start
	fmt_done
	dev_idle
*/ 

#if DEBUGGING > 2

pid_t	floppy_event(int Event)
{
	if (Event < 0 || Event >= MAX_EVENT) {
		ll_debug();
	}
	if (FdcState < 0 || FdcState >= MAX_STATE) {
		ll_debug();
	}
disp_str("change:");
disp_hex(FdcState);
disp_hex(Event);
	return  (*(action[FdcState][Event]))();
}
#elif DEBUGGING
pid_t	floppy_event(int Event)
{
	VPUTC(FdcState+'0');
	VPUTC(Event+'A');
	return (*(action[FdcState][Event]))();
}
#else 

#define	floppy_event(evnt)	((*(action[FdcState][evnt]))())

#endif




long	io_sequence = 0;

/*
 *  Driver_io
 *
 *  Starts the io request, returning to the filesystem.  At some later point
 *  in time, the request completes and the filesystem is awoken (proxied).
 *  Driver_io must make a copy of it's parameters, start (or schedule) the
 *  io request, the return.  If it encounters any immediate errors, can
 *  abandon the request.
 *
*/

int __far drvio(unsigned disknum, unsigned mode,
          unsigned nblocks, unsigned rsvd,
          struct _block __far *bp)
{
DISK	*dp;
int		rblock;

	if (rsvd)
		;
	io_sequence++;

	if (bp->b_block == 1) {
		disp_str("block 0 read!\n");
	}
#if DEBUGGING > 1
disp_str("IOREQUEST=");	disp_lhex(bp->b_block); disp_hex(nblocks);
#endif

	if (disknum > MAX_NDRIVE) {
		return -1;			/*	not even close */
	}
	if ((dp=disk + disknum)->d_busy) {
		ll_debug();
		return -1;
	}
	if (!(FdcState == ST_IDLE || FdcState == ST_START)) {
		if (FdcState != ST_READY) {	/*	bad news! */
			return -1;
		}
		if (DriveNo != disknum) {
			FdcState = ST_IDLE;
		}
	}
/*	SetUp residual parameters to execute command */

	dp->d_busy = 1;
	Function = mode;
	CachePtr = bp;
	CurRequest.block = FirstBlock = bp->b_block  - 1;
	map_thn(&CurRequest,dp);

	cur_disk = DriveParms + (DriveNo = disknum);
	

/*	
	Don't allow Multi-Sector IO at root.
	Limit it to the number of blocks on a cylinder

	rblock is the block number relative to this head.
*/

	rblock = dp->d_sctr_trk * CurRequest.head + CurRequest.sector;
        
	if (multi_track) {
		if (CurRequest.block == 0) {
			NumBlocks = 1;			/*	force this */
		} else if (rblock+nblocks > dp->d_sctr_cyl) {
			NumBlocks = dp->d_sctr_cyl - rblock + 1;
		} else {
			NumBlocks = nblocks;
		}   
	} else {
		if (CurRequest.block == 0) {
			NumBlocks = 1;
		} else if (rblock + nblocks > dp->d_sctr_trk) {
			NumBlocks = dp->d_sctr_trk - rblock + 1;
		} else {
			NumBlocks = nblocks;
		}
	}

	if (mode == WRITE_BLK) {			
		/*	squirrel it away while we can address the cache buffer! */
		cache_to_buf(iobuffer.bufp, bp, NumBlocks);
	}
	return floppy_event(IORequest) ? -1 : 0;
}

/*
 * Allows ASCommands to be sent down into device.

 */

#define	BIT_WIDTH	010
#define	BITS_SHIFT	003
#define	BITS_MASK	007
#define	_THIS_BYTE(_i)	((_i) >> BITS_SHIFT)
#define	_THIS_BIT(_i)	(1 << ((_i)&BITS_MASK))
#define	ADDSET(_X,_i)	((_X)[_THIS_BYTE(_i)] |=  _THIS_BIT(_i))



int __far drvctl(unsigned disknum, unsigned request,
				 unsigned rsvd1, unsigned rsvd2,
				 void __far *cdata)
{
FLOPPY_PARMS	*p;
union {
	struct	_qio_getcap	get_cap;
	struct	_sysmsg_version_reply vers;
	struct	_qio_cap		cap;
	struct	fd_parms		drv_parm;
	struct  fd_phys_parms	drv_phys;
	struct	fd_fmt_trk		fmt_track;
} __far *parm;

	if (rsvd1 || rsvd2)
		;
	disp_str("QNX_IOCTL "); disp_hex(request);

	parm = cdata;

	switch (request) {
	case	QCTL_COMM_GET_CAP:
		if (parm->get_cap.start_num == QCTL_FSYS) {
		int	i;
			for (i=0; i < sizeof parm->cap.qctl_vect; i++)
				parm->cap.qctl_vect[i] = 0;
			ADDSET(parm->cap.qctl_vect, QCTL_GET_DISK_PARMS - QCTL_COMM_GET_CAP);
			ADDSET(parm->cap.qctl_vect, QCTL_SET_DISK_PARMS - QCTL_COMM_GET_CAP);
			ADDSET(parm->cap.qctl_vect, QCTL_GET_PHYS_PARMS - QCTL_COMM_GET_CAP);
			ADDSET(parm->cap.qctl_vect, QCTL_SET_PHYS_PARMS - QCTL_COMM_GET_CAP);
			ADDSET(parm->cap.qctl_vect, QCTL_FMT_TRK - QCTL_COMM_GET_CAP);
			parm->cap.len = sizeof(parm->cap.qctl_vect);
			return 0;
		} else {
			parm->cap.len = -1;
			return -1;
		}
		break;
	case	QCTL_GET_DISK_PARMS:
		parm->drv_parm.type 	= disk[disknum].d_type;
		parm->drv_parm.subtype	= DriveParms[disknum].type; 
		parm->drv_parm.heads	= disk[disknum].d_heads;
		parm->drv_parm.tracks	= disk[disknum].d_cyls;
		parm->drv_parm.sectors	= disk[disknum].d_sctr_trk;
		parm->drv_parm.blocksize= _BLOCK_SIZE;
		parm->drv_parm.blocks	= disk[disknum].d_num_blks;
		return 0;
		break;

	case	QCTL_SET_DISK_PARMS:
		disk[disknum].d_type	= parm->drv_parm.type; 
		DriveParms[disknum].type= parm->drv_parm.subtype; 
		disk[disknum].d_heads	= parm->drv_parm.heads; 
		disk[disknum].d_cyls	= parm->drv_parm.tracks; 
		disk[disknum].d_sctr_trk= parm->drv_parm.sectors; 

/*
 *		_BLOCK_SIZE				= parm->drv_parm.blocksize; 
 */
		disk[disknum].d_num_blks= parm->drv_parm.blocks; 
		return 0;
		break;		

	case	QCTL_GET_PHYS_PARMS:
		
		p = DriveParms + disknum;
		parm->drv_phys.dma_channel	= dma_channel;
		parm->drv_phys.read_gap		= p->rd_gap[p->cur];
		parm->drv_phys.step_rate	= p->st_rate[p->cur];
		parm->drv_phys.shift_step	= p->dstep[p->cur];
		parm->drv_phys.pll_clock	= p->clock[p->cur];
		parm->drv_phys.head_load_time= p->head_load;
		parm->drv_phys.head_unload_time= p->head_unload;
		parm->drv_phys.format_gap    = p->fmt_gap[p->cur];
		return 0;
		break;		

	case	QCTL_SET_PHYS_PARMS:

		p = DriveParms + disknum;
		dma_channel			= parm->drv_phys.dma_channel; 
		p->rd_gap[p->cur]	= parm->drv_phys.read_gap; 
		p->st_rate[p->cur]	= parm->drv_phys.step_rate; 
		p->dstep[p->cur]	= parm->drv_phys.shift_step; 
		p->clock[p->cur]	= parm->drv_phys.pll_clock; 
		p->head_load		= parm->drv_phys.head_load_time; 
		p->head_unload		= parm->drv_phys.head_unload_time; 
		p->fmt_gap[p->cur]	= parm->drv_phys.format_gap; 
		return 0;
		break;		

	case	QCTL_FMT_TRK:
		if (disk[disknum].d_busy) {	/*	look out... here comes trouble */
			ll_debug();
		}
		disk[disknum].d_busy = 1;
		disk[disknum].d_nblks = 1;
		disk[disknum].d_retval = 0;
		
		fmttrk = &(parm->fmt_track);
		cur_disk = DriveParms + (DriveNo = disknum);
		Function = FMT_TRACK;
		CurRequest.head = fmttrk->head;
		CurRequest.track = fmttrk->trackno;
		return floppy_event(IORequest) ? -1 : 0;
		break;

	default:
		return -1;
		break;
	}
	return 0;
}

int __far drvctmask(unsigned disknum, unsigned request,
				unsigned rsvd1, unsigned rsvd2)
{
int		iflags = _DRVR_CTRL_SUPPORTED;

	if (disknum || rsvd1 || rsvd2)
		;
	switch (request) {
	case	QCTL_COMM_GET_CAP:
		break;
	case	QCTL_GET_DISK_PARMS:
		break;
	case	QCTL_SET_DISK_PARMS:
		iflags |= _DRVR_CTRL_PRIVILIGED | _DRVR_CTRL_RECONFIG;
		break;		
	case	QCTL_GET_PHYS_PARMS:
		break;		
	case	QCTL_SET_PHYS_PARMS:
		iflags |= _DRVR_CTRL_PRIVILIGED | _DRVR_CTRL_RECONFIG;
		break;		
		break;		
	case	QCTL_FMT_TRK:
		iflags |= _DRVR_CTRL_PRIVILIGED | 
					_DRVR_CTRL_RECONFIG |
					_DRVR_CTRL_DESTRUCTIVE |
					_DRVR_CTRL_ASYNCH;
		break;
	default:
		iflags = 0;
		break;
	}
	return iflags;
}

LOCAL	volatile int	timer_active = 0;
LOCAL	volatile int	timer_count = 0;

#define	kill_wakeup()	( timer_active = 0 )
#define	enable_wakeup()	( timer_active = 1 )

pid_t	__far	int_timer()
{
	if (!timer_active)
		return 0;	
	--timer_count;
	if (timer_count == 0) {
		kill_wakeup();
		return floppy_event(TimeOut);
	}
	return 0;
}

static	int	set_wakeup(int ticks)
{
int	psw;
	psw = disable();
		kill_wakeup();
		timer_count = ticks;
		enable_wakeup();
	restore(psw);
	return 0;
}



/*
 * int_handler
 *  This routine is executed at interrupt time.
 *  When the current io request is completed (failed or success) it
 *  must awake the filesystem|itself to signal io completion.
 *  If it wakes the filesystem, it must update the appropriate structures
 *  including success code.
 *	Only have addressing into his cache while in his ldt.
 */





/*	

	When the handler 'sees' a block 0 fly by, squirrel away it's 
	parameter table into this structure.  Another level of software 
	can determine what to do with it.

*/

#define	MAGIC_LOCATION	(512-18)

#define	QNX_2_MAGIC_NUM	0x3412
#define	QNX_4_MAGIC_NUM 0xbeba
#define	QNX_4OLD_MAGIC_NUM 0xcefa


pid_t __far	int_handler()
{
	return floppy_event(Interrupt);
}



remount_disk()
{
unsigned char	__far	*bufp = iobuffer.bufp;
unsigned	p0, p1, p2;

/*	first try qnx 4.0 style disk */
	if (*(unsigned __far *)(bufp+512-16) == QNX_4_MAGIC_NUM) {
		bufp += 512-16;
		p0 = *(unsigned __far *)(bufp +8);
		p1 = *(unsigned __far *)(bufp +12);
		p2 = *(unsigned __far *)(bufp +10);
		if (p0 == 0 || p1 == 0 || p2 == 0)
			return 1;
		disk[DriveNo].d_heads = p0;
		disk[DriveNo].d_cyls =  p1;
		disk[DriveNo].d_sctr_trk = p2;
		disk[DriveNo].d_sctr_cyl = p2 * p0;
		disk[DriveNo].d_num_blks = p0 * p2 * p1;
		return 0;
	}

/*	then qnx 2.1 style disk */
	else if (*(unsigned __far *)(bufp+512-18) == QNX_2_MAGIC_NUM) {
		bufp += 512 - 18;

		p0 =	*(bufp+10);		
		p1 =	*(unsigned __far *)(bufp+8);
		p2 =	*(unsigned __far *) (bufp+6);
		
		if (p0 == 0 || p1 == 0 || p2 == 0) 
			return 1;

		disk[DriveNo].d_sctr_trk =	p0;		
		disk[DriveNo].d_sctr_cyl =	p1;
		disk[DriveNo].d_num_blks =	p2;
		disk[DriveNo].d_heads = 
			disk[DriveNo].d_sctr_cyl/disk[DriveNo].d_sctr_trk;
		disk[DriveNo].d_cyls = 
			disk[DriveNo].d_num_blks/disk[DriveNo].d_sctr_cyl;

		return 0;
	}
/* then a dos one. */	
	switch (bufp[0x15]) {
	case	0xf0:
	case	0xf8: 	case	0xf9:
	case	0xfa: 	case	0xfb:
	case	0xfc: 	case	0xfd:
	case	0xfe: 	case	0xff:

		p0 = *(unsigned __far *)(bufp+24);	/*	sectors	*/
		p1 = *(unsigned __far *)(bufp+26);	/*	heads	*/
		p2 = *(unsigned __far *)(bufp+19);	/*	blocks */
		if (p0 == 0 || p1 == 0 || p2 == 0)
			return 1;
		disk[DriveNo].d_sctr_trk = p0;
		disk[DriveNo].d_heads    = p1;
		disk[DriveNo].d_num_blks = p2;
		disk[DriveNo].d_sctr_cyl = p0 * p1;
		disk[DriveNo].d_cyls = 
	 					disk[DriveNo].d_num_blks / disk[DriveNo].d_sctr_cyl;
		return 0;
		break;
	default:
		return 1;
	}

}


/*
 * Allways update state variables before calling finish_io.
 */

pid_t finish_io(int last_block, int errcode)
{

	disk[DriveNo].d_nblks = last_block - FirstBlock;

	if (errcode != _DRVR_OK) {
#if DEBUGGING > 1
disp_str("finish_io: error code =");
disp_hex(errcode);
#endif
	}

/*	change this when there is a new filesystem */
	if (errcode == _DRVR_OK)
		disk[DriveNo].d_retval = (disk[DriveNo].d_nblks == NumBlocks) 		
								?	 0	:  _DRVR_PART_IO;
	else
		disk[DriveNo].d_retval = errcode;
	
#if DEBUGGING > 1
disp_str("NBLOCKS="); disp_hex(disk[DriveNo].d_nblks);
#endif
	disk[DriveNo].d_busy = 0;
	return NotifyPid;
}

ioctl_done()
{
	VPUTC('+');
	VPUTC(NotifyPid);
	disk[DriveNo].d_retval = 0;
	disk[DriveNo].d_busy = 0;
	return NotifyPid;
}


/*
 * after many attempts to organize these routines, I have finally
 * come upon this order:
 *	first routines called in response to an INT
 *	second, routines called in response to a TICK
 *  third, routines called by IO requests.
 */
LOCAL	int	retry_count = 0;

pid_t	absorb_int()
{
	retry_count = 0;
	switch (fdc_direction()) {
	case	0:	/* no RQM */
		break;
	case	1:
		fdc_status();
		break;
	case	2:
		setup_sense_intr();
		fdc_cmd(0);
		break;
	}
	return 0;
}

pid_t	init_dev()
{
	kill_wakeup();
	if (fdc_direction() == 1)
		fdc_status();
	setup_sense_intr(); 
	FdcState = ST_IDLE;
	if (fdc_cmd(0) < 0)	{
		VPUTC('@');
		ll_debug();
	}
	return init_io();
}



pid_t	io_start()
{
int	seek_stat;

	if (fdc_direction() == 1)
		fdc_status();

	seek_stat = seek_ok();

	/*	we're ok! */
	switch (Function) {
	case	READ_BLK:
	case	WRITE_BLK:
		if (seek_stat != 0) {
			if (++retry_count < MAX_RETRY) {
				return seek_drive();
			}
			set_wakeup(SPINDOWN);
			FdcState = ST_READY;
			return finish_io(FirstBlock,_DRVR_DISK_FAILURE);
		}
		FdcState = ST_XFER;
		setup_dma(dma_flags | ((Function == WRITE_BLK) ? DMA_OUTPUT : 0),
			&iobuffer, NumBlocks * _BLOCK_SIZE, dma_channel);
		setup_xfer(Function, DriveNo,
				CurRequest.head, CurRequest.track, CurRequest.sector,
				disk[DriveNo].d_sctr_trk, cur_disk->rd_gap[cur_disk->cur]);
		break;
	case	FMT_TRACK:
		FdcState = ST_FMT_TRK;
		setup_dma(dma_flags | DMA_OUTPUT, &iobuffer, 
				fmttrk->nsector * 4, dma_channel);
		setup_fmt_trk(DriveNo, fmttrk,
					cur_disk->fmt_gap[cur_disk->cur], iobuffer.bufp);
		break;
	}
	set_wakeup(XFER_TIMEOUT);
	if (fdc_cmd(1) < 0) {
		VPUTC('$');
		ll_debug();
	}
	return 0;
}




pid_t	io_done()
{
int		ind;
long	lblock;
unsigned	rcode = _DRVR_OK;

	kill_wakeup();
	if ((ind = fdc_status()) == 0) {
		lblock = FirstBlock + NumBlocks;
    } else if (fdc_xfer_ok()) {
		lblock = fdc_last_block(disk[DriveNo].d_sctr_cyl,
							disk[DriveNo].d_sctr_trk)
					- disk[DriveNo].d_sctr_base;
		rcode = _DRVR_PART_IO;
	} else {
		lblock = FirstBlock;
		rcode = _DRVR_ERROR;
	}
	if (lblock > FirstBlock) {
		if (FirstBlock == 0) {
			remount_disk();
		}
		FdcState = ST_READY;
		if (Function == READ_BLK) {
			buf_to_cache(CachePtr, iobuffer.bufp, lblock - FirstBlock);
		}
		set_wakeup(WAIT_SPINDOWN);
		return finish_io(lblock,rcode);
	}
	if (++retry_count >= MAX_RETRY) {
		FdcState = ST_READY;
		set_wakeup(SPINDOWN);
		return finish_io(FirstBlock,rcode);
	}
	switch (retry_count) {
	case	2:
	case	4:
		if (FirstBlock == 0) {		
	   		cur_disk->cur = ! cur_disk->cur;		/*	switch to next parms */
			fdc_setclock(cur_disk->clock[cur_disk->cur]);
			fdc_specify(cur_disk->st_rate[cur_disk->cur],
				cur_disk->head_unload,
				cur_disk->head_load);
	case	1: /* careful ... */	
			return seek_drive();
		}	/* fall-through */
	case	3:
		FdcState = ST_SPECIFY;
		setup_recalibrate(DriveNo);
		if (fdc_cmd(1) < 0) {
			VPUTC('%');
			ll_debug();
		}
		break;
	}
	return 0;
}




pid_t	retry()
{
	if (fdc_direction() == 1) 
		fdc_status();

	if (seek_ok() == 0) {	/*	recalibrate took! */
		return seek_drive();
	}
	FdcState = ST_SPECIFY2;
	setup_recalibrate(DriveNo);
	if  (fdc_cmd(1) < 0) {
		VPUTC('^');
		ll_debug();
	}
	return 0;
}
pid_t	retry2()
{
	if (fdc_direction() == 1)
		fdc_status();
	if (seek_ok() == 0) {	/*	recalibrate took! */
		return seek_drive();
	}
/*	An error has occured, go look at the Result Phase. */
#if DEBUGGING
		ll_debug();
#else
	FdcState = ST_READY;
	set_wakeup(SPINDOWN);
	return finish_io(FirstBlock,_DRVR_DISK_FAILURE);
#endif
}

pid_t	fmt_done()
{
	kill_wakeup();
	FdcState = ST_READY;
	if (fdc_direction() == 1)
		fdc_status();
	if (fdc_fmt_ok()) {
		nec_setres((struct fd_last_err __far *)fmttrk);
		disp_str("Disk Formatting Error!\n");
#if DEBUGGING
		disp_Result();
#endif
	} else {
		((struct fd_last_err __far *)fmttrk)->result = FD_OK;
		((struct fd_last_err __far *)fmttrk)->head = CurRequest.head;
		((struct fd_last_err __far *)fmttrk)->track = CurRequest.track;
		((struct fd_last_err __far *)fmttrk)->sector = 0;
	}
	set_wakeup(WAIT_SPINDOWN);
	return ioctl_done();
}



pid_t	reset_dev()
{
	retry_count= 0;
	set_wakeup(2);
	FdcState = ST_RESET;
	fdc_reset();
	return 0;
}


pid_t	init_io()
{
int	timeout = READ_SPINUP;

	if (Function != READ_BLK)
		timeout = WRITE_SPINUP;


	FdcState = ST_SPIN;
	fdc_motor_on(DriveNo);
	fdc_setclock(cur_disk->clock[cur_disk->cur]);
	fdc_specify(cur_disk->st_rate[cur_disk->cur],
			cur_disk->head_unload,
			cur_disk->head_load);

	set_wakeup(timeout);
	retry_count = 0;
	return 0;
}

pid_t	ready_io()
{
	kill_wakeup();
	retry_count = 0;
	return seek_drive();
}

pid_t	seek_drive()
{
	setup_seek(DriveNo, CurRequest.track, cur_disk->dstep[cur_disk->cur]);
	FdcState = ST_SEEK;
	if (fdc_cmd(1) < 0) {
		VPUTC('#');
		ll_debug();
	}
	return 0;
}


pid_t	dev_idle()
{
	FdcState = ST_IDLE;
	fdc_motor_off(DriveNo);
	return 0;
}


pid_t	xfer_to()
{
	FdcState = ST_START;
	clear_dma(dma_flags, dma_channel);		/*	stop the xfer */
	fdc_motor_off(DriveNo);					/* motor is shut down */
	return finish_io(FirstBlock, _DRVR_DISK_FAILURE);
}


pid_t	fmt_to()
{
/*	reset controller -- don't know where it's been */
	FdcState = ST_START; 
	clear_dma(dma_flags, dma_channel);		/*	stop the xfer */

	((struct fd_last_err __far *)fmttrk)->result = FD_NOT_READY;
	((struct fd_last_err __far *)fmttrk)->head = CurRequest.head;
	((struct fd_last_err __far *)fmttrk)->track = CurRequest.track;
	((struct fd_last_err __far *)fmttrk)->sector = 0;

	fdc_motor_off(DriveNo);
	return ioctl_done();
}



pid_t	io_trap()	/*	unexpected IO Request. */
{
	ll_debug();
	FdcState = ST_RTRAP;		/*	IO Request Trap */
	clear_dma(dma_flags, dma_channel);
	return 0;
}

pid_t	timer_trap()	/*	unexpected Timer Expiry */
{
	ll_debug();
	FdcState = ST_TTRAP;
	clear_dma(dma_flags, dma_channel);
	return 0;
}

pid_t	int_trap()		/*		unexpected interrupt */
{
	ll_debug();
	if (fdc_direction())
		fdc_status();
	FdcState = ST_ITRAP;
	clear_dma(dma_flags, dma_channel);
	return 0;
}


get_floppy_parms(DISK *dp, int disknum)
{
int	status;
FLOPPY_PARMS	*fp = DriveParms+disknum;
	status =  disknum 
				?	(get_cmos(0x010) & 0x0f)
				:	(get_cmos(0x010) >> 4);
	if (status == 0 || status > 4) {
		dp->d_type = _UNMOUNTED;
		return -1;
	} 
	dp->d_type = _FLOPPY;
	dp->d_sctr_base = 1;

	fp->type = status;
	fp->locked = fp->cur = 0;
	fp->head_load   = 0x07;
	fp->head_unload = 0x0f;

	switch (status) {
	case	1:		/*	360  */
		dp->d_cyls     = 40;
		dp->d_heads    = 2;
		dp->d_sctr_trk = 9;

		fp->clock[0]   = fp->clock[1]   = FD_300KBS;
		fp->st_rate[0] = fp->st_rate[1] = 0x0d;
		fp->rd_gap[0]  = fp->rd_gap[1]  = 0x2a;
		fp->dstep[0]   = fp->dstep[1]   = 0;
		fp->fmt_gap[0] = fp->fmt_gap[1] = 0x2f;
		break;

	case	2:		/*	1.2  */
		dp->d_cyls     = 80;
		dp->d_heads    = 2;
		dp->d_sctr_trk = 15;

		fp->clock[0] 	= FD_500KBS;		fp->clock[1] = FD_300KBS;	
		fp->st_rate[0]	= 0x0a; 			fp->st_rate[1] = 0x0c;		
		fp->rd_gap[0]	= 0x1b; 			fp->rd_gap[1] = 0x23;		
		fp->dstep[0]	= 0; 				fp->dstep[1] = 1;			
		fp->fmt_gap[0]	= 0x6c; 			fp->fmt_gap[1] = 0x2f; 
		break;

	case	3:		/*	720  */
	
		dp->d_cyls     = 80;
		dp->d_heads    = 2;
		dp->d_sctr_trk = 9;

		fp->clock[0]   = fp->clock[1]   = FD_250KBS;
		fp->st_rate[0] = fp->st_rate[1] = 0x0d;
		fp->rd_gap[0]  = fp->rd_gap[1]  = 0x2a;
		fp->dstep[0]   = fp->dstep[1]   = 0;
		fp->fmt_gap[0] = fp->fmt_gap[1] = 0x2f;
		break;

	case	4:		/*	1.44	*/
		dp->d_cyls     = 80;
		dp->d_heads    = 2;
		dp->d_sctr_trk = 18;

		fp->clock[0]	= FD_500KBS; 	fp->clock[1] = FD_250KBS;
		fp->st_rate[0]	= 0x0a; 		fp->st_rate[1] = 0x0d;
		fp->rd_gap[0]	= 0x1b; 		fp->rd_gap[1] = 0x2a;
		fp->dstep[0]	= 0;			fp->dstep[1] = 0;
		fp->fmt_gap[0]	= 0x6c;			fp->fmt_gap[1] = 0x2f;
		break;
	}

	dp->d_sctr_cyl = dp->d_sctr_trk * dp->d_heads;
	dp->d_num_blks = dp->d_sctr_cyl * dp->d_cyls;	

	return status;
}
