#ifdef __USAGE
%C - control a tape driver (QNX)

%C	[-n Name] Command [Command...]
Options:
 -n         Names the block special device to use.
Where:
     Name is:  /dev/tp0   (by default)

  Command is:  bot        Move to beginning of tape.
               eod        Move to end of data.
               erase      Move to bot and erase.
               forward    Move forward to next file mark.
               log        Display driver log history.
               query      Read & Print the driver status.
               reset      Reset the tape controller (status MUST follow).
               retension  Move bot, eot then bot.
               rewind     (same as bot)
               status     Read & Print the current tape status information.
               version    Read & Print the current driver version information.
               writefm    Write a file mark at the current position.
#endif
/*
**	tape
**
**	This utility allows the manipulation of tape drives through
**	the use of IOCNTL commands sent through Fsys.
**
**	It is expected that this utility would be used in
**	conjunction with archiver utilities such as pax, tar or
**	cpio.
**
**	A couple of commands will also allow the user to query the
**	driver to determine how many read/write errors occurred to
**	get a feel for the reliability of the tape and tape drive. 
**	An abnormally high number of errors may mean that the tape
**	in use is marginal or that the drives ability to write to
**	the tape is in some way impeeded (i.e. dirty heads).
**
**	Written by Rob Oakley.
**
**	May 6, 1991 - Utility created to manipulate quarter inch cartridge tape
**	              conforming to the jello QIC 02 standard.
**	Jan 2, 1992 - Added support for message log.  Removed support for
**				  debug log.
**
*/
#include	<ctype.h>
#include	<errno.h>
#include	<fcntl.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<unistd.h>

#include	<sys/disk.h>
#include	<sys/fsys.h>
#include	<sys/qioctl.h>
#include	<sys/stat.h>
#include	<sys/sys_msg.h>
#include	<sys/types.h>

#ifndef		TRUE
#	define	TRUE	1
#	define	FALSE	0
#endif

/*
**	Local Private Variables
*/
static char					*ProgramNamep = "$Id: tape.c,v 1.4 1992/10/27 19:05:25 eric Exp $";
static char					 TapeFullPath[ sizeof "/dev/" + _DRIVER_NAME_LEN + 1 ] = "/dev/tp0";
static QIC02_MSG_STRUCT		 qic02ms;


/*****************************************************************/
static void			PrintMessageLog( int fd, QIC02_MSG_STRUCT *qic02msp )
/*****************************************************************/
{	if ( *qic02msp->Body.Log.Msg )
		while ( 1 )
		{	fprintf( stdout, "%s: %s\n", ProgramNamep, qic02msp->Body.Log.Msg );

			if ( qic02msp->Body.Log.Again )
			{	if ( qnx_ioctl(	fd,
								QCTL_RAW_CMD,
								qic02msp, sizeof (QIC02_HEADER_STRUCT),
								qic02msp, sizeof (QIC02_MSG_STRUCT) ) == -1 )
				{	fprintf( stderr, "%s: Unable to Issue command to Tape\n", ProgramNamep );
					exit( 7 );
				}
			}
			else
				break;
		}
	else
		fprintf( stdout, "%s: Log is empty.\n", ProgramNamep );
}


/*****************************************************************/
static void			 PrintStatus( QIC02_MSG_STRUCT *qic02msp, int err_and_underrun )
/*****************************************************************/
{
	int				 i;
	unsigned int	 s = (unsigned int) qic02msp->Body.Status.Byte1 << 8 | qic02msp->Body.Status.Byte0;

	static char		*StatusStr[] =
	{	"File Mark Detected",
		"Block Not Located",
		"Unrecoverable Data Error",
		"End Of Tape",
		"Write Protect Cartridge",
		"Device Fault or SELF TEST Failed",
		"Cartridge Not in Place",
		NULL,
		"Power Up/RESET Occurred",
		"End of Recorded Data",
		"Bus Parity Error",
		"Beginning of Tape",
		"Marginal Block Detected",
		"No Data Detected",
		"Illegal Command",
		NULL
	};

	if ( qic02msp->Header.RetVal == EOK )
	{	for ( i = 0;  s;  i++, s >>= 1 )
			if ( s & 0x0001  &&  StatusStr[i] )	
				fprintf( stdout, "%s: Unit Reports \"%s\"\n", ProgramNamep, StatusStr[ i ] );

		if ( err_and_underrun )
		{	fprintf( stdout, "%s:\t   Error Count = %u\n", ProgramNamep, (unsigned int) qic02msp->Body.Status.ErrorCounth		<< 8 | qic02msp->Body.Status.ErrorCountl );
			fprintf( stdout, "%s:\tUnderrun Count = %u\n", ProgramNamep, (unsigned int) qic02msp->Body.Status.UnderRunCounth	<< 8 | qic02msp->Body.Status.UnderRunCountl );
		}
	}
	else
		fprintf( stdout, "%s: Unable to read unit status, check log\n", ProgramNamep );
}

/*****************************************************************/
static void			ReadAndPrintStatus( int fd )
/*****************************************************************/
{	QIC02_MSG_STRUCT	qic02ms;

	qic02ms.Header.Command = T_READ_STATUS;

	if ( qnx_ioctl(	fd,
					QCTL_RAW_CMD,
					&qic02ms,
					sizeof (QIC02_HEADER_STRUCT),
					&qic02ms,
					sizeof (QIC02_LAST_ERROR_STRUCT) ) == -1 )
		fprintf( stdout, "%s: Unable to Read Status\n", ProgramNamep );
	else
		PrintStatus( &qic02ms, FALSE );
}

/*****************************************************************/
void				PrintVersion( QIC02_MSG_STRUCT *qic02mp )
/*****************************************************************/
{	fprintf(	stdout, "%s:  %s - %s, Version %d.%d%c\n",
				ProgramNamep,
				qic02mp->Body.Version.name,
				qic02mp->Body.Version.date,
				qic02mp->Body.Version.version / 100,
				qic02mp->Body.Version.version - (qic02mp->Body.Version.version / 100 * 100),
				qic02mp->Body.Version.letter
			);
}


/*****************************************************************/
					main( int argc, char *argv[] )
/*****************************************************************/
{
	int				  i,
					  numberofcommands,
					  invalidcommand = FALSE;
	unsigned int	 *cp,
					 *commandp;
	char			**argvpp;
	int				  fd;

	extern char		 *optarg;
	extern int		  optind,
					  opterr,
					  optopt;


	ProgramNamep = argv[0];

	/*
	**	Process command line options
	*/
	opterr = 0;
	while ( (i = getopt( argc, argv, "n:" )) != -1 )
		switch( i )
		{	case 'n':
				strncpy( TapeFullPath, optarg, sizeof (TapeFullPath) - 1 );
			break;

			default:
				fprintf( stderr, "%s: Invalid Option '%c'\n", ProgramNamep, optopt );
				exit( 1 );
			break;
		}
	
	/*
	**	Open the tape unit
	*/
	if ( (fd = open( TapeFullPath, O_WRONLY )) == -1 )
	{	fprintf( stderr, "%s: ERROR Unable to Open %s, errno is \"%s\"\n", ProgramNamep, TapeFullPath, strerror( errno ) );
		exit( 2 );
	}

	if ( (numberofcommands = argc - optind) == 0 )
	{	fprintf( stderr, "%s: ERROR Command Not Specified\n", ProgramNamep );
		exit( 3 );
	}

	if ( (cp = malloc( numberofcommands * sizeof (qic02ms.Header.Command) )) == NULL )
	{	fprintf( stderr, "%s: ERROR Unable to allocate %d bytes of memory\n", ProgramNamep, numberofcommands * sizeof (unsigned int) );
		exit( 4 );
	}

	i			= optind;
	argvpp		= &argv[ i ];
	commandp	= cp;

	/*
	**	Convert commands to their unsigned int equivalents
	*/
	while ( i < argc )
	{	switch ( tolower( **argvpp ) )
		{	case 'b':
				if ( strcmpi( "bot", *argvpp ) == 0 )
					*commandp++ = T_BOT;
				else
					invalidcommand = TRUE;
			break;

			case 'e':
				if ( strcmpi( "eod", *argvpp ) == 0 )
					*commandp++ = T_SEEK_EOD;
				else if ( strcmpi( "erase", *argvpp ) == 0 )
					*commandp++ = T_ERASE;
				else
					invalidcommand = TRUE;
			break;

			case 'f':
				if ( strcmpi( "forward", *argvpp ) == 0 )
					*commandp++ = T_SEEK_FM;
				else
					invalidcommand = TRUE;
			break;

			case 'l':
				if ( strcmpi( "log", *argvpp ) == 0 )
					*commandp++ = T_DEBUG;
				else
					invalidcommand = TRUE;
			break;

			case 'q':
				if ( strcmpi( "query", *argvpp ) == 0 )
					*commandp++ = T_LAST_ERROR;
				else
					invalidcommand = TRUE;
			break;

			case 'r':
				if ( strcmpi( "reset", *argvpp ) == 0 )
					*commandp++ = T_RESET;
				else if ( strcmpi( "retension", *argvpp ) == 0 )
					*commandp++ = T_RETENSION;
				else if ( strcmpi( "rewind", *argvpp ) == 0 )
					*commandp++ = T_BOT;
				else
					invalidcommand = TRUE;
			break;

			case 's':
				if ( strcmpi( "status", *argvpp ) == 0 )
					*commandp++ = T_READ_STATUS;
				else
					invalidcommand = TRUE;
			break;

			case 'v':
				if ( strcmpi( "version", *argvpp ) == 0 )
					*commandp++ = T_VERSION;
				else
					invalidcommand = TRUE;
			break;

			case 'w':
				if ( strcmpi( "writefm", *argvpp ) == 0 )
					*commandp++ = T_WRITE_FM;
				else
					invalidcommand = TRUE;
			break;

			default:
				invalidcommand = TRUE;
		}

		if ( invalidcommand )
		{	fprintf( stderr, "%s: ERROR Invalid Command \"%s\"\n", ProgramNamep, *argvpp );
			exit( 5 );
		}

		i++;
		argvpp++;
	}

	i			= optind;
	commandp	= cp;

	while ( i < argc )
	{	qic02ms.Header.Command = *commandp;

		if ( qnx_ioctl(	fd,
					QCTL_RAW_CMD,
					&qic02ms, sizeof (QIC02_HEADER_STRUCT),
					&qic02ms, sizeof (QIC02_MSG_STRUCT) ) == -1 )
		{	fprintf( stderr, "%s: Unable to Issue \"%s\" to Tape.  Use `tape log' for detail.\n", ProgramNamep, argv[i] );
			exit( 6 );
		}

		switch ( *commandp )
			{	case T_READ_STATUS:	PrintStatus( &qic02ms, TRUE );	break;
				case T_DEBUG:		PrintMessageLog( fd,&qic02ms );	break;
				case T_VERSION:		PrintVersion( &qic02ms );		break;

				case T_LAST_ERROR:
					fprintf( stdout, "%s: Driver Reports \"%s\"\n", ProgramNamep, qic02ms.Body.Error.ErrorString );
				break;

				default:
					if ( qic02ms.Header.RetVal != EOK )
						ReadAndPrintStatus( fd );
			}

		i++;
		commandp++;
	}

	close( fd );
	free( cp );

	exit( 0 );
}


/* End of File */
